/* name: kr6.java
 * description: interacts with kr6.wrl in order to fix the position of the arms
 *  -reads datagrams from port 4849 (generated by posserver)
 *  -writes on port 4848 to instant player (virtual arm)
 * author: tilman jochems
 * (c)27/05/2010, Midi-Formatique, Rivesaltes */

import java.io.*;
import java.net.*;
import java.util.*;


//class for manipulation of kr6
public class kr6{

 //constants
 static final int EAIPORT=4848;						//vrml viewer
 static final int POSPORT=4849;						//socket to posserver
 static final String WRLARMNAME="ARM";				//name of arms in wrl file

 //nested class PosRead : thread for reading values
 private static PosReader Pos=null;
 static class PosReader extends Thread {

  //public variables
  boolean NewData=false;								//new data read
  float[] Axis=new float[6];							//result from posserver

  //private variables
  private byte[] Frame=new byte[64];			  	//for datagram
  private DatagramSocket Socket;
  private DatagramPacket Packet;

  //construction
  PosReader() {
   try{
    //create socket
    Socket=new DatagramSocket(POSPORT);
    Packet=new DatagramPacket(Frame,Frame.length);
   }//try
   catch (Throwable all){							  //stop application if error
    all.printStackTrace();
    Socket=null;
    System.exit(0);
   }//catch
  }//construction

  //free socket
  public void finalize(){
   Socket.close();
  }//finalize

  //the thread main loop
  public void run(){
   while(true){

    try{
     //receive the next packet (blocking)
     Socket.receive(Packet);

     //convert string to individual float values (in degree)
     synchronized(this){
      StringTokenizer st=new StringTokenizer(new String(Frame));
      for(int i1=0;i1<6;i1++) Axis[i1]=Float.valueOf(st.nextToken());
      NewData=true;									//flag: we have data
     }//synchronized
    }//try
    catch (Throwable all){}			 			//just continue reading

   }//while
  }//run
 }//class PosReader


 //main function (entry point of class kr6)
 //loops by reading position from PosReader and writing to EAI interface
 public static void main(String[] args){
  vrml.eai.Browser browser=null;

  try{
   int i1;
	long l1;
   int af=11;												//actuel frame counter
   long ltime=0; 	   									//time of last read
   float[][] rot=new float[6][];			   		//rotation information
   float[]  lpos=new float[6];	 	 	 			//position of axis
	float[]  npos=new float[6];
	float[]  apos=new float[6];

   vrml.eai.field.EventInSFRotation set_rot[]=new vrml.eai.field.EventInSFRotation[6];

   //say hello
   System.out.println("MFJ Arm Server");

   //initialize positions
   for(i1=0;i1<6;i1++) apos[i1]=npos[i1]=lpos[i1]=0;

	//try to get link to the arms of the robot (15s)
	java.net.InetAddress address=java.net.InetAddress.getByName("localhost");
   l1=System.currentTimeMillis();
   do{
    try{
     browser=vrml.eai.BrowserFactory.getBrowser(address,EAIPORT);
     browser.setDescription("KR6/2");

     //get rotation fields
     vrml.eai.Node arm;
     for(i1=0;i1<6;i1++){
      arm=browser.getNode(WRLARMNAME + (i1+1));
      set_rot[i1]=(vrml.eai.field.EventInSFRotation)arm.getEventIn("set_rotation");
     }//for

     break;
    }//try
    catch(Throwable all){Thread.sleep(100);}
   }while(System.currentTimeMillis()-l1<15000);
   if(browser==null) throw new Error("can't connect to the vrml viewer");

   //add a listener to the browser
   browser.addBrowserListener(
    new vrml.eai.event.BrowserListener(){
	  public void browserChanged(vrml.eai.event.BrowserEvent evt){
	   kr6.onBrowserChanged(evt);
    }});

   //create and start posreader
   Pos=new PosReader();
   Pos.start();

   //the never ending story
   while(true){

    //is there a new position?
    if(Pos.NewData){

     //get new axis position and transform radian
     synchronized(Pos){
      for(i1=0;i1<6;i1++)npos[i1]=Pos.Axis[i1]*(float)Math.PI/180;
      Pos.NewData=false;
     }//synchronized

     //copy actuel position to last
     for(i1=0;i1<6;i1++) lpos[i1]=apos[i1];

     //start visualisation sequence
     af=0;
     ltime=System.currentTimeMillis();
    }//if NewData

    //every new value will be displayed with higher frequence
    if(af>=10){
     Thread.sleep(20);									//no new data

    }else{
     //calculate intermediate approximated position
     af++;
     for(i1=0;i1<6;i1++)apos[i1]=lpos[i1]+(npos[i1]-lpos[i1])*af/10;

     //set rotation information (depends on wrl model)
     rot[0]=new float[]{0,-1,0,apos[0]};
     rot[1]=new float[]{0,0,-1,apos[1]};
     rot[2]=new float[]{0,0,-1,apos[2]};
     rot[3]=new float[]{-1,0,0,apos[3]};
     rot[4]=new float[]{0,0,-1,apos[4]};
     rot[5]=new float[]{-1,0,0,apos[5]};

     //update scene
     browser.beginUpdate();
     set_rot[0].setValue(rot[0]);
     set_rot[1].setValue(rot[1]);
     set_rot[2].setValue(rot[2]);
     set_rot[3].setValue(rot[3]);
     set_rot[4].setValue(rot[4]);
     set_rot[5].setValue(rot[5]);
     browser.endUpdate();

     //timing: show 10 frames in 200ms
     l1=ltime+af*20-System.currentTimeMillis();
     if(l1>=1 && l1<=100) Thread.sleep(l1);
     //System.out.print(af + " " + l1 + ", ");
    }//if
   }//while
  }//try
  catch (Throwable all){
   all.printStackTrace();
  }
  finally{
   if (browser!=null)browser.dispose();
  }
 }//main

 //exit application on browser communication error
 private static void onBrowserChanged(vrml.eai.event.BrowserEvent evt){

   //close system if communication with browser brakes
  switch (evt.getID()){
   case vrml.eai.event.BrowserEvent.INITIALIZED: break;
   case vrml.eai.event.BrowserEvent.SHUTDOWN:
   case vrml.eai.event.BrowserEvent.URL_ERROR:
   case vrml.eai.event.BrowserEvent.CONNECTION_ERROR:
   default:
    Pos=null;												//end thread
    System.exit(0);
  }//switch
 }//onBrowerChanged
}//class
